// Map.cc

#include "Map.h"
#include "SFMT.h"
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <string>

int Map::map_width, Map::map_height;
uint64_t Map::map_mask;
uint64_t Map::hashkeys_moving;
uint64_t* Map::hashkeys_walls;
uint64_t* Map::hashkeys_bikes;

Map::Map() {
  ReadFromFile(stdin);
}

Map::Map(const Map& other)
{
    moving = other.moving;
    bikes[0][0] = other.bikes[0][0];
    bikes[0][1] = other.bikes[0][1];
    bikes[1][0] = other.bikes[1][0];
    bikes[1][1] = other.bikes[1][1];
    hash = other.hash;
    space = Territory::allocate(*(other.space));
}

Map::~Map()
{
    Territory::free(space);
}

Map& Map::operator=(const Map& other)
{
    moving = other.moving;
    bikes[0][0] = other.bikes[0][0];
    bikes[0][1] = other.bikes[0][1];
    bikes[1][0] = other.bikes[1][0];
    bikes[1][1] = other.bikes[1][1];
    hash = other.hash;
    (*space) = (*other.space);
    return *this;
}

bool Map::operator==(const Map& other) const
{
    if (this == &other)
        return true;
    if (hash != other.hash
            || moving != other.moving)
        return false;
    if (bikes[0][0] != other.bikes[0][0]
            || bikes[0][1] != other.bikes[0][1]
            || bikes[1][0] != other.bikes[1][0]
            || bikes[1][1] != other.bikes[1][1])
        return false;
    if ((*space) != (*other.space))
        return false;
    return true;
}

bool Map::isWall(const int* const cord) const {
  if (cord[0] < 0 || cord[1] < 0
          || cord[0] >= map_width || cord[1] >= map_height) {
      return true;
  } else {
      return ((*space)[cord[0]] & (UINT64_C(1) << cord[1])) == 0;
  }
}

bool Map::isWall(int x, int y) const
{
    if (x < 0 || y < 0 || x >= map_width || y >= map_height)
    {
        return true;
    }
    return ((*space)[x] & (UINT64_C(1) << y)) == 0;
}

std::string& Map::toString()
{
    std::string* board = new std::string();
    for (int j=0; j < map_height; j++)
    {
        for (int i=0; i < map_width; i++)
        {
            bool at_bike1 = bikes[0][0] == i && bikes[0][1] == j;
            bool at_bike2 = bikes[1][0] == i && bikes[1][1] == j;

            if (at_bike1 && !at_bike2)
            {
                if (isWall(i, j))
                    board->append("!");
                else
                    board->append("1");
            }
            else if (at_bike2 && !at_bike1)
            {
                if (isWall(i, j))
                    board->append("@");
                else
                    board->append("2");
            }
            else if (at_bike1 && at_bike2)
            {
                board->append("B");
            }
            else if (isWall(i, j))
            {
                board->append("#");
            } else {
                board->append(" ");
            }
        }
        board->append("\n");
    }
    return *board;
}

void Map::makeMove(int move) {
  fprintf(stdout, "%d\n", move+1);
  fflush(stdout);
}

void Map::doPlayoutStep(int direction)
{
    int* cur_bike = bikes[moving];
    (*space)[cur_bike[0]] &= ~(UINT64_C(1) << cur_bike[1]) & map_mask;
    switch (direction)
    {
        case 0:
            cur_bike[1] -= 1;
            break;
        case 1:
            cur_bike[0] += 1;
            break;
        case 2:
            cur_bike[1] += 1;
            break;
        case 3:
            cur_bike[0] -= 1;
            break;
        default:
            fprintf(stderr, "Bad direction given to doPlayoutStep\n");
            break;
    }
    moving ^= 1;
}

void Map::doStep(int direction)
{
    int startx = bikes[moving][0];
    int starty = bikes[moving][1];
    hash ^= hashkeys_bikes[(moving * map_width * map_height)
        + (startx * map_height) + starty];
    (*space)[startx] &= ~(UINT64_C(1) << starty) & map_mask;
    hash ^= hashkeys_walls[(startx * map_height) + starty];
    switch(direction)
    {
        case 0:
            bikes[moving][1] -= 1;
            break;
        case 1:
            bikes[moving][0] += 1;
            break;
        case 2:
            bikes[moving][1] += 1;
            break;
        case 3:
            bikes[moving][0] -= 1;
            break;
    }
    int finalx = bikes[moving][0];
    int finaly = bikes[moving][1];
    hash ^= hashkeys_bikes[(moving * map_width * map_height)
        + (finalx * map_height) + finaly];
    moving = moving ^ 1;
    hash ^= hashkeys_moving;
}

void Map::undoStep(int direction)
{
    moving = moving ^ 1;
    hash ^= hashkeys_moving;
    int startx = bikes[moving][0];
    int starty = bikes[moving][1];
    hash ^= hashkeys_bikes[(moving * map_width * map_height)
        + (startx * map_height) + starty];
    switch(direction)
    {
        case 0:
            bikes[moving][1] += 1;
            break;
        case 1:
            bikes[moving][0] -= 1;
            break;
        case 2:
            bikes[moving][1] -= 1;
            break;
        case 3:
            bikes[moving][0] += 1;
            break;
    }
    int finalx = bikes[moving][0];
    int finaly = bikes[moving][1];
    (*space)[finalx] |= UINT64_C(1) << finaly;
    hash ^= hashkeys_walls[(finalx * map_height) + finaly];
    hash ^= hashkeys_bikes[(moving * map_width * map_height)
        + (finalx * map_height) + finaly];
}

void Map::move_bike(int direction)
{
    int startx = bikes[0][0];
    int starty = bikes[0][1];
    hash ^= hashkeys_bikes[(startx * map_height) + starty];
    (*space)[startx] &= ~(UINT64_C(1) << starty) & map_mask;
    hash ^= hashkeys_walls[(startx * map_height) + starty];
    switch(direction)
    {
        case 0:
            bikes[0][1] -= 1;
            break;
        case 1:
            bikes[0][0] += 1;
            break;
        case 2:
            bikes[0][1] += 1;
            break;
        case 3:
            bikes[0][0] -= 1;
            break;
    }
    int finalx = bikes[0][0];
    int finaly = bikes[0][1];
    hash ^= hashkeys_bikes[(finalx * map_height) + finaly];
}

void Map::reset_bike(int direction)
{
    int startx = bikes[0][0];
    int starty = bikes[0][1];
    hash ^= hashkeys_bikes[(startx * map_height) + starty];
    switch(direction)
    {
        case 0:
            bikes[0][1] += 1;
            break;
        case 1:
            bikes[0][0] -= 1;
            break;
        case 2:
            bikes[0][1] -= 1;
            break;
        case 3:
            bikes[0][0] += 1;
            break;
    }
    int finalx = bikes[0][0];
    int finaly = bikes[0][1];
    (*space)[finalx] |= UINT64_C(1) << finaly;
    hash ^= hashkeys_walls[(finalx * map_height) + finaly];
    hash ^= hashkeys_bikes[(finalx * map_height) + finaly];
}

bool Map::isEnd() const
{
    if (moving)
        return false;
    if (bikes[0][0] == bikes[1][0] && bikes[0][1] == bikes[1][1])
        return true;
    if (!((*space)[bikes[0][0]] & (UINT64_C(1) << bikes[0][1]))
            || !((*space)[bikes[1][0]] & (UINT64_C(1) << bikes[1][1])))
        return true;
    return false;
}

int Map::score() const
{
    if (moving)
        return 0;
    bool myCrash = ((*space)[bikes[0][0]]
            & (UINT64_C(1) << bikes[0][1])) == 0;
    bool oppCrash = ((*space)[bikes[1][0]]
            & (UINT64_C(1) << bikes[1][1])) == 0;
    if (myCrash && !oppCrash)
        return -1;
    if (oppCrash && !myCrash)
        return 1;
    return 0;
}

void Map::ReadFromFile(FILE *file_handle) {
  int x, y, c;
  int width, height;
  int num_items = fscanf(file_handle, "%d %d\n", &width, &height);
  if (feof(file_handle) || num_items < 2) {
    exit(0); // End of stream means end of game. Just exit.
  }
  if (map_width != width || map_height != height)
  {
      if (map_width || map_height)
      {
          fprintf(stderr, "Map size changed from previous size\n");
          exit(1);
      }
      map_width = width;
      map_height = height;
      map_mask = (UINT64_C(1) << height) - 1;
      generateKeys();
      Territory::preallocate();
  }
  space = Territory::allocate();
  moving = 0;
  x = 0;
  y = 0;
  while (y < map_height && (c = fgetc(file_handle)) != EOF) {
    switch (c) {
    case '\r':
      break;
    case '\n':
      if (x != map_width) {
        fprintf(stderr, "x != width in Board_ReadFromStream\n");
        return;
      }
      ++y;
      x = 0;
      break;
    case '#':
      if (x >= map_width) {
        fprintf(stderr, "x >= width in Board_ReadFromStream\n");
        return;
      }
      (*space)[x] &= ~(UINT64_C(1) << y);
      ++x;
      break;
    case ' ':
      if (x >= map_width) {
        fprintf(stderr, "x >= width in Board_ReadFromStream\n");
        return;
      }
      (*space)[x] |= UINT64_C(1) << y;
      ++x;
      break;
    case '1':
      if (x >= map_width) {
        fprintf(stderr, "x >= width in Board_ReadFromStream\n");
        return;
      }
      (*space)[x] |= UINT64_C(1) << y;
      bikes[0][0] = x;
      bikes[0][1] = y;
      ++x;
      break;
    case '2':
      if (x >= map_width) {
        fprintf(stderr, "x >= width in Board_ReadFromStream\n");
        return;
      }
      (*space)[x] |= UINT64_C(1) << y;
      bikes[1][0] = x;
      bikes[1][1] = y;
      ++x;
      break;
    default:
      fprintf(stderr, "unexpected character %d in Board_ReadFromStream", c);
      return;
    }
  }
  generateHash();
}

void Map::generateHash()
{
    hash = 0;
    for (int i=0; i < map_width; i++)
    {
        for (int j=0; j < map_height; j++)
        {
            if (!((*space)[i] & (UINT64_C(1) << j)))
                hash ^= hashkeys_walls[(i * map_height) + j];
        }
    }
    for (int i=0; i < 2; i++)
    {
        int x = bikes[i][0];
        int y = bikes[i][1];
        hash ^= hashkeys_bikes[(i * map_width * map_height)
            + (x*map_height) + y];
    }
    if (moving)
        hash ^= hashkeys_moving;
}

void Map::generateKeys()
{
    hashkeys_moving = gen_rand64();
    hashkeys_walls = new uint64_t[map_width * map_height];
    for (int i=0; i < (map_width * map_height); i++)
    {
        hashkeys_walls[i] = gen_rand64();
    }
    hashkeys_bikes = new uint64_t[2 * map_width * map_height];
    for (int i=0; i < (2 * map_width * map_height); i++)
    {
        hashkeys_bikes[i] = gen_rand64();
    }
}
