// Map.h
//
// Handles the Tron map. Also handles communicating with the Tron game engine.
// You don't need to change anything in this file.

#define __STDC_CONSTANT_MACROS

#include "territory.h"
#include <string>

#pragma once

enum Bound {
    LOWER,
    EXACT,
    UPPER
};

class Map {
 public:
  // Constructs a Map by reading an ASCII representation from the console
  // (stdin).
  Map();

  // Copy contructor
  Map(const Map& other);

  ~Map();

  // Returns whether or not the given cell is a wall or not. TRUE means it's
  // a wall, FALSE means it's not a wall, and is passable. Any spaces that are
  // not on the board are deemed to be walls.
  bool isWall(const int* const cord) const;
  bool isWall(int x, int y) const;

  std::string& toString();

  // Sends your move to the contest engine. The four possible moves are
  //   * 1 -- North. Negative Y direction.
  //   * 2 -- East. Positive X direction.
  //   * 3 -- South. Positive X direction.
  //   * 4 -- West. Negative X direction.
  static void makeMove(int move);

  // Make a step without updating the hash
  void doPlayoutStep(int direction);

  // Make a step
  void doStep(int direction);

  // Undo a step
  void undoStep(int direction);

  void move_bike(int direction);
  void reset_bike(int direction);

  bool isEnd() const;
  int score() const;

  void generateHash();

  Map& operator=(const Map& other);
  bool operator==(const Map& other) const;
  bool operator!=(const Map& other) const { return !(*this == other); }

  // Map dimensions.
  static int map_width, map_height;
  static uint64_t map_mask;

  // The locations of both players.
  int bikes[2][2];

  // side moving
  int moving;

  // position hash
  uint64_t hash;

  Territory* space;

  static uint64_t* hashkeys_walls;
  static uint64_t* hashkeys_bikes;

 private:
  // Load a board from an open file handle. To read from the console, pass
  // stdin, which is actually a (FILE*).
  //   file_handle -- an open file handle from which to read.
  //
  // If there is a problem, the function returns NULL. Otherwise, a valid
  // Board structure is returned.
  //
  // The file should be an ascii file. The first line contains the width and
  // height of the board, separated by a space. subsequent lines contain visual
  // representations of the rows of the board, using '#' and space characters.
  // The starting positions of the two players are indicated by '1' and '2'
  // characters. There must be exactly one '1' character and one '2' character
  // on the board. For example:
  // 6 4
  // ######
  // #1# 2#
  // #   ##
  // ######
  void ReadFromFile(FILE *file_handle);

  static void generateKeys();
  static void preallocate();

  static uint64_t hashkeys_moving;
};
