
#include "territory.h"
#include "Map.h"
#include <vector>

#pragma once

namespace cnct
{

class Node;

class ChildInfo
{
    public:
    Node* node;
    int exit[2];
    int best_size;
    Bound bound;

    ChildInfo(Node* n) : best_size(0), node(n) { }
};

class Node
{
    public:
    static Node* allocate();
    static Node* free(Node* node);

    void clear();

    int age;
    Node* parent;
    int entrance[2];
    bool doubled_entrance;
    Territory* area;
    std::vector<ChildInfo> children;

    int my_size;
    Bound my_bound;

    int set_size();

#ifdef MEMORY_CHECK
    static int allocated;
    static int unreturned();
#endif
    private:
    static std::vector<Node*> cache;
    Node();
    ~Node();
};

class Tree
{
    public:
    Node* root;
    Territory* area;
    Node*** node_map;

    Tree(Territory& terr, int* start);
    ~Tree();

    std::string& toString();

#ifdef MEMORY_CHECK
    static int allocated;
    static int unreturned();
#endif
    private:
    static std::vector<Node***> map_cache;
    static Node*** allocate_map();
    static Node*** free_map(Node*** map);

    Node* combine_nodes(Node* first, Node* second);
    void collapse_doubled(Node* node);

    int age;
};

class TreeSizeCache
{
    public:
    TreeSizeCache();
    int get_size(Territory& area, int* start, Bound& bound);

    private:
    uint64_t* hashes;
    int* sizes;
    Bound* bounds;
};

}
