
#include "time_util.h"

#if defined(_MSC_VER)

#include <Windows.h>

uint64_t perf_freq;

void init_timer()
{
    LARGE_INTEGER li;
    QueryPerformanceFrequency(&li);
    perf_freq = li.QuadPart;
}

uint64_t millis()
{
    LARGE_INTEGER count;
    QueryPerformanceCounter(&count);
    return count.QuadPart / (perf_freq / 1000);
}

#else

#include <sys/time.h>

void init_timer() { }

uint64_t millis()
{
    timeval now;
    uint64_t millis;
    gettimeofday(&now, 0);
    millis = now.tv_sec * 1000;
    millis += now.tv_usec / 1000;
    return millis;
}

#endif

void StopWatch::set_length(int milliseconds)
{
    search_time = milliseconds;
}

void StopWatch::start()
{
    start_time = millis();
    nodes_visited = 0;
    next_check = 1000;
    stop_time = start_time + search_time;
    last_time = start_time;
    last_check = 0;
}

uint64_t StopWatch::elapsed()
{
    return millis() - start_time;
}

bool StopWatch::check_abort()
{
    uint64_t now = millis();
    if (now >= stop_time)
        return true;
    uint64_t check_time = now - last_time;
    double short_npms, long_npms;
    if (check_time < 5)
        short_npms = (nodes_visited - last_check) / (double)(check_time + 2);
    else
        short_npms = (nodes_visited - last_check) / (double)check_time;
    uint64_t search_time = now - start_time;
    if (start_time < 5)
        long_npms = nodes_visited / (double)(search_time + 2);
    else
        long_npms = nodes_visited / (double)search_time;
    double npms = short_npms < long_npms ? short_npms : long_npms;
    int left = (int)(stop_time - now);
    if (left > 150)
    {
        next_check = (int)(npms * 100) + nodes_visited;
    }
    else if (left > 50)
    {
        next_check = (int)((npms * (left-25)) / 2) + nodes_visited;
    }
    else if (left > 10)
    {
        next_check = (int)((npms * left) / 4) + nodes_visited;
    }
    else
    {
        next_check = (int)(npms * 2) + nodes_visited + 1;
    }
    last_check = nodes_visited;
    last_time = now;
    return false;
}


